/* --COPYRIGHT--,BSD
 * Copyright (c) 2016,2019, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

/********************************************************************************
 *
 *  stepper.c - Definition file for functions related to stepper motion control
 *  DRV8889-Q1_Example FIRMWARE
 *  1/2/2019
 *
 *******************************************************************************/

/********************************************************************************
 * HEADER FILES
 *******************************************************************************/
#include <stepper.h>

/********************************************************************************
 * MACROS
 *******************************************************************************/
#pragma DATA_SECTION(gStepper_Obj,".stpVarSection");
volatile Stepper_Driving_Obj_t gStepper_Obj;

/*****************************************************************************
* Define Global Variables in stepper.h
*****************************************************************************/
//#define DISABLE_STALL_DURING_ACCEL
/********************************************************************************
 * LOCAL ROUTINES
 *******************************************************************************/
/*
 * TimerA Initialization for PWM
 * From MSP430x261x_ta_13.c in MSPWARE
 */
void stepper_Init(void)
{

    set_DRVOFF_hi;

    // configure timer A
    TACTL =  CNTL_0 | TASSEL_2 |ID_0 | MC_0;

    // configure timer A
    // select SMCLK, divide by 1, stop
    TBCTL =  TBCLGRP_0 | TBSSEL_2 |ID_3 | MC_0;

    P1SEL |= STEP;                            // STEP pin GPIO select

    // Set Default GUI Variables Values
    SetDefaultGUIVals();

    gStepper_Obj.gDIR = false;

    // Load GPIO Values
    UpdateDeviceControls();
    UpdateGUIControls();

    // set boolean defaults
    LastCommandWasRunSteps = false;
    MoveSteps = false;
    UpdateSteppingRateTMR = false;
    temp_reciprocation  = false;
    RecoverFromStall = false;
    stall_learn_number_of_steps = 0;
    stall_learn_microstep_divider = 0;
    last_stall_state = DISABLED;
    SPI_STEP_DIR_CMD = 0;
    SteppedOnce = 3;
}

void SetDefaultGUIVals()
{
    last_reciprocation_state = 0;
//    stall_fault_detected = 0;

    gStepper_Obj.MOTOR_CMD_ACTIVE1 = false;
    gStepper_Obj.MOTOR_CMD_ACTIVE2 = false;
    gStepper_Obj.MOTOR_CMD_ACTIVE3 = false;
    gStepper_Obj.SPI_STEP_DIR_EN = false;
    gStepper_Obj.STARTING_SPEED = DEFAULT_START_SPEED;     // initial speed in pulses per second when starting motor
    gStepper_Obj.TARGET_SPEED = DEFAULT_TARGET_SPEED;      // speed of motor once running
    gStepper_Obj.ACCEL_RATE = DEFAULT_ACCEL_RATE;          // acceleration rate from starting speed to target speed
    gStepper_Obj.STOPPING_SPEED = DEFAULT_STOPPING_SPEED;  // speed to stop motor once deceleration has reached this point
    gStepper_Obj.NUM_STEPS = DEFAULT_NUM_STEPS;            // number of steps to move when manually stepping
    gStepper_Obj.STEPS_2_STOP = DEFAULT_STEPS_2_STOP;      // number of steps used to stop the motor when manually stepping
    gStepper_Obj.IFS = DEFAULT_IFS;
    IFS_previous = DEFAULT_IFS+1;
    gStepper_Obj.REVERSE_ON_STALL = false;

    gStepper_Obj.RUN_STEPPER = false;                // Auto run enabled
    gStepper_Obj.MANUAL_ADVANCE = false;           // Manually advance motor
    gStepper_Obj.RECIPROCATION = false;

    gStepper_Obj.StallState = DISABLED;
}

//
// routine to communicate between GUI and device
// reads GUI variables and updates the GPIOs

void UpdateDeviceControls()
{
    // toggle status if count is reached
    if (LED_TOGGLE_CNT == LED_TIMEOUT)
    {
        toggle_status_led;
        LED_TOGGLE_CNT = 0;
    }

    // From GUI to Micro
    // Update the GPIO Pins from Global Variables

    /* During Fault Condition*/
    // From Micro to GUI
    if (read_nFAULT_pin == 0)
        gDRV8889_Obj.nFault = 0;
    else
        gDRV8889_Obj.nFault = 1;

    if (gDRV8889_Obj.nSleep == 1)
        set_nSLEEP_hi;
    else
        set_nSLEEP_lo;

    // DIR
    if (gStepper_Obj.gDIR == 1)
        set_DIR_hi;
    else
        set_DIR_lo;

    // ENABLE  -- set port high to enable device
    if (gDRV8889_Obj.gDRVOFF == 0)
        set_DRVOFF_lo;
    else
        set_DRVOFF_hi;
}

//
// routine to communicate between device and GUI
// reads GPIOs and updates the GUI

void UpdateGUIControls()
{
    // Enable Motor control buttons
    if (gDRV8889_Obj.gDRVOFF == 1 || gDRV8889_Obj.nSleep == 0)
    {
        // when disabled reset the indicators in the motor operation
        gStepper_Obj.RUN_STEPPER = false;
        gStepper_Obj.MANUAL_ADVANCE = false;
        gStepper_Obj.RECIPROCATION = false;
        gStepper_Obj.MOTOR_CMD_ACTIVE1 = false;
        gStepper_Obj.MOTOR_CMD_ACTIVE2 = false;
        gStepper_Obj.MOTOR_CMD_ACTIVE3 = false;
        // stop the timer also
        TACTL &=  ~MC_3;                    // Stop timer by setting mc_0
        TACCR0 = 0x0000;                    // once stopped reset the timer value
        // clear temp_reciprocation and MoveSteps so next movement is correct
        temp_reciprocation = false;         // reset reciprocation
        MoveSteps = false;
        gStepper_Obj.AccelerateState = IDLE; // set AccelerateState to idle when
                    // sleeping or disabled
    }
    if (gStepper_Obj.IFS != IFS_previous)
    {
        DAC12_1DAT = 2730*gStepper_Obj.IFS;
        IFS_previous = gStepper_Obj.IFS;
        _nop();
    }
}

//
// routine to determine if any changes are in motor operation are required
// and pass the information to the motor state machine
//
void UpdateSpeed()
{
    StartingSpeed = (unsigned long) gStepper_Obj.STARTING_SPEED << use_one_edge;       // Configure the starting speed
    StoppingSpeed = (unsigned long) gStepper_Obj.STOPPING_SPEED << use_one_edge;       // Configure the stopping speed
    StepsToMove   = (unsigned long) gStepper_Obj.NUM_STEPS << use_one_edge;            // adjust for both edges
    StepsToStop   = (unsigned long) gStepper_Obj.STEPS_2_STOP << use_one_edge;         // adjust for both edges
    DesiredTargetSpeed = (unsigned long) gStepper_Obj.TARGET_SPEED << use_one_edge;    // Configure the target speed
    AccelRate     = (unsigned long) gStepper_Obj.ACCEL_RATE << use_one_edge;           // Configure the acceleration rate here

    DetermineMotorMovement();                               // Once type of motor movement is determined
    MotorStateMachine();                                    // Command the motor to move
}

//
//  Routine to determine desired motor operation
//  direction, acceleration, deceleration, steps to move
//  reciprocation
//
void  DetermineMotorMovement()
{
    // calculate steps to target and steps to stop prior to starting motor
    if ((gStepper_Obj.RUN_STEPPER == true || gStepper_Obj.MANUAL_ADVANCE == true || gStepper_Obj.RECIPROCATION == true) && \
         gStepper_Obj.MOTOR_CMD_ACTIVE2 == false && gStepper_Obj.MOTOR_CMD_ACTIVE1 == false && gStepper_Obj.MOTOR_CMD_ACTIVE3 == false )
    {
        // update steps to speed and steps to stop calculations
        // only calculate when the motor is idle and commanded to move
        time = (float)(DesiredTargetSpeed - StartingSpeed)/AccelRate;
        time_squared = time * time;
        // calculate steps to target
        Steps2Target = ((AccelRate >> 1) * (time_squared)) + (StartingSpeed * time);
        // calculate Steps2Stop if:
        //  StepsToStop in GUI is >1 and
        //  StoppingSpeed is not equal Starting Speed
        //
        if ((StepsToStop == 2 && use_one_edge == 1) || StepsToStop == 1)
        {
            if (StartingSpeed != StoppingSpeed)
            {   // recalculate time if necessary
                time = (float)(DesiredTargetSpeed - StoppingSpeed)/AccelRate;
                time_squared = time * time;
            }
            // calculate steps required to stop
                Steps2Stop = ((AccelRate >> 1) * (time_squared)) + (StartingSpeed * time);
        }
        else    // defined number of steps to stop
        {
            Steps2Stop = StepsToStop;
        }
        // calculate stepping rate update (integer value divide by 32)
        if (AccelRate > 31 || AccelRate == 0)                   // if >=32 or 0
        {
            SteppingRateUpdate = AccelRate >> 5;                // divide by 32 - right shift 5
        }
        else
        {
            SteppingRateUpdate = 1;                             // increment by 1 to get correct value
        }
        if ((StartingSpeed >= DesiredTargetSpeed) || (StoppingSpeed >= DesiredTargetSpeed))
            // Special case -- override prior settings
            // allow motor to attempt to run only at starting speed
        {
            SteppingRateUpdate = 0;
            Steps2Target = 0;
            Steps2Stop = 0;
        }
        // transition from ACCEL to DECEL when count equals (StepsToMove - Steps2Stop) <=0
        if ((Steps2Stop > (StepsToMove >>1 )) && (StartingSpeed == StoppingSpeed))
            Steps2Decel = StepsToMove >>1;                      // decel at midpoint
        else
            Steps2Decel = StepsToMove - Steps2Stop;
    }
    // start running or stopping the motor if commanded and not manually advancing
    if (gStepper_Obj.RUN_STEPPER == true && gStepper_Obj.MOTOR_CMD_ACTIVE1 == false && gStepper_Obj.MOTOR_CMD_ACTIVE2 == false && gStepper_Obj.MOTOR_CMD_ACTIVE3 == false)
    {
        // START STEPPER
        gStepper_Obj.MOTOR_CMD_ACTIVE1 = true;                 // Indicate that the motor is running

        LastCommandWasRunSteps = true;              // flag to determine if the timer requires clearing

        DesiredStepperSpeed = DesiredTargetSpeed;   // Configure the final stepping rate in PPS

        SteppingRate = StartingSpeed;               // This is the starting speed from the GUI
        // adjusted by the edge indicator
        SteppingRateTMR = SMCLK_FREQ/StartingSpeed; // Convert here to the timer value -- xxTMR means the
        // value to be loaded into the timer
        gStepper_Obj.AccelerateState = ACCEL;                    // signal to start state machine
        // Configure TimerA
        TACCTL1 &= 0xFF1F;                          // Clear OUTMODx bits; 3 MSB on lower byte; Timer configured to output mode
        TACCTL1 |= OUTMOD_4 + CCIE;                 // Configure the timer as TOGGLE. TBCCR0 does the reset
        TACCR0 = SteppingRateTMR;                   // Start the timer here with initial value
        TACCR1 = SteppingRateTMR >> 1;              // set TBCCR1 to the midpoint of the TBCCR0 for toggle
        // configure TimerB
        TBCCTL1 &= 0xFF1F;                          // Clear OUTMODx bits; 3 MSB on lower byte; Timer configured to output mode
        TBCCTL1 |= OUTMOD_4 + CCIE;
        TBCCR0 = TIME_32ms;                         // set timer to 32ms
        TBCCR1 = TIME_32ms >> 1;
        // start timers
        TACTL |= MC_1;                              // Start timerA by selecting up mode
        TBCTL |= MC_1;                              // start timerB
#ifdef DISABLE_STALL_DURING_ACCEL
        if(gStepper_Obj.REVERSE_ON_STALL == true)   // disable stall detection during acceleration
        {
            SPI_writeRegister(SPI_REG_CTRL5, DIS_STL_MASK);
        }
#endif
    }
    if ((gStepper_Obj.RUN_STEPPER == false && gStepper_Obj.MOTOR_CMD_ACTIVE1 == true))
        // motor is running and stop is issued
    {
        // STOP STEPPER
        gStepper_Obj.MOTOR_CMD_ACTIVE1 = false;                // must set low when stopping
        gStepper_Obj.AccelerateState = DECEL;
        DesiredStepperSpeed = StoppingSpeed;        //Speed to stop at
    }
    // MOVE_STEPS
    // Manually advance steps if commanded
    // only allow starting if not automatically running
    if (gStepper_Obj.MANUAL_ADVANCE == true && gStepper_Obj.MOTOR_CMD_ACTIVE1 == false && gStepper_Obj.MOTOR_CMD_ACTIVE2 == false && gStepper_Obj.MOTOR_CMD_ACTIVE3 == false)
    {
        gStepper_Obj.MOTOR_CMD_ACTIVE2 = true;
        if (LastCommandWasRunSteps == true || use_one_edge == true)         // 15APR -- added control to avoid resetting
        {
            // reset timer only when starting or when using one edge
            TACTL |= TACLR;                         // reset timerA
            TACTL = CNTL_0 | TASSEL_2 |ID_0 | MC_0;
            TACCTL1 &= 0xFF1F;                      // Clear OUTMODx bits; 3 MSB on lower byte; Timer configured to output mode
            TACCTL1 |= OUTMOD_4 + CCIE;             // Configure the timer as SET_RESET. TBCCR0 does the reset
        }
        LastCommandWasRunSteps = false;             // reset flag to indicate move steps was last commanded
        DesiredStepperSpeed = DesiredTargetSpeed;   // Configure the Frequency Rate

        SteppingRate = StartingSpeed;
        SteppingRateTMR = SMCLK_FREQ/StartingSpeed; // convert to timer value
        gStepper_Obj.AccelerateState = ACCEL;                    // signal to start state machine
        tmpStepsToMove = 0;                         // clear counter
        MoveSteps = true;                           // set flag to indicate manual advance
        // configure Timer A
        TACCR0 = SteppingRateTMR;                   // start the timer
        TACCR1 = SteppingRateTMR >> 1;              // now using count up
        // configure TimerB
        TBCCTL1 &= 0xFF1F;                          // Clear OUTMODx bits; 3 MSB on lower byte; Timer configured to output mode
        TBCCTL1 |= OUTMOD_4 + CCIE;
        TBCCR0 = TIME_32ms;                         // set timer to 32ms
        TBCCR1 = TIME_32ms >> 1;
        // start timers
        TACTL |= MC_1;                              // Start timer by selecting up mode
        TBCTL |= MC_1;
#ifdef DISABLE_STALL_DURING_ACCEL
        if(gStepper_Obj.REVERSE_ON_STALL == true)   // disable stall detection during acceleration
        {
            SPI_writeRegister(SPI_REG_CTRL5, DIS_STL_MASK);
        }
#endif
    }

    if(gStepper_Obj.RECIPROCATION == true && gStepper_Obj.MANUAL_ADVANCE == true && temp_reciprocation == false)
    {
        last_reciprocation_state = 1;
        temp_reciprocation = true;
        gStepper_Obj.MOTOR_CMD_ACTIVE3 = true;
        if (LastCommandWasRunSteps == true || use_one_edge == true)         // 15APR -- added control to avoid resetting
        {
            // reset timer only when starting or when using one edge
            TACTL |= TACLR;                             // reset timerA
            TACTL = CNTL_0 | TASSEL_2 |ID_0 | MC_0;
            TACCTL1 &= 0xFF1F;                          // Clear OUTMODx bits; 3 MSB on lower byte; Timer configured to output mode
            TACCTL1 |= OUTMOD_4 + CCIE;                 // Configure the timer as SET_RESET. TBCCR0 does the reset
        }
            LastCommandWasRunSteps = false;             // reset flag to indicate move steps was last commanded
            DesiredStepperSpeed = DesiredTargetSpeed;   // Configure the Frequency Rate

            SteppingRate = StartingSpeed;
            SteppingRateTMR = SMCLK_FREQ/StartingSpeed; // convert to timer value
            __delay_cycles(10000000);
            gStepper_Obj.AccelerateState = ACCEL;       // signal to start state machine
            tmpStepsToMove = 0;                         // clear counter
            MoveSteps = true;                           // set flag to indicate manual advance
            // configure Timer A
            TACCR0 = SteppingRateTMR;                   // start the timer
            TACCR1 = SteppingRateTMR >> 1;              // now using count up
            // configure TimerB
            TBCCTL1 &= 0xFF1F;                          // Clear OUTMODx bits; 3 MSB on lower byte; Timer configured to output mode
            TBCCTL1 |= OUTMOD_4 + CCIE;
            TBCCR0 = TIME_32ms;                         // set timer to 32ms
            TBCCR1 = TIME_32ms >> 1;
            // start timers
            TACTL |= MC_1;                              // Start timer by selecting up mode
            TBCTL |= MC_1;
#ifdef DISABLE_STALL_DURING_ACCEL
        if(gStepper_Obj.REVERSE_ON_STALL == true)   // disable stall detection during acceleration
        {
            SPI_writeRegister(SPI_REG_CTRL5, DIS_STL_MASK);
        }
#endif
    }
    // Special case of updating speed during running
    // if the Desired Target Speed has been updated, then the stepper speed should be increased or decreased
    //
    if ((gStepper_Obj.RUN_STEPPER == true) && (DesiredStepperSpeed != DesiredTargetSpeed))
    {
        if (DesiredTargetSpeed > DesiredStepperSpeed)
        {
            gStepper_Obj.AccelerateState = ACCEL;
#ifdef DISABLE_STALL_DURING_ACCEL
            if(gStepper_Obj.REVERSE_ON_STALL == true)   // disable stall detection during acceleration
            {
                SPI_writeRegister(SPI_REG_CTRL5, DIS_STL_MASK);
            }
#endif
        }
        else
            gStepper_Obj.AccelerateState = DECEL2NEWTARGET;
        // reset DesiredStepperSpeed
        DesiredStepperSpeed = DesiredTargetSpeed;
    }
    // Special case of stall fault while driving.
    // Reverse direction
    //if (read_nFAULT_pin == 0 && (gStepper_Obj.RUN_STEPPER == true) && (gStepper_Obj.REVERSE_ON_STALL == true) && (SPI_readRegister(SPI_REG_DIAG_STATUS_2) & 0x08))
    if (read_nFAULT_pin == 0 && gStepper_Obj.StallState != DISABLED && (gStepper_Obj.REVERSE_ON_STALL == true) && (SPI_readRegister(SPI_REG_DIAG_STATUS_2) & 0x08))
    {
        RecoverFromStall = true;
        gStepper_Obj.AccelerateState = IDLE;
        last_stall_state = gStepper_Obj.StallState;
        if(gStepper_Obj.StallState == READY4STALL && (SPI_readRegister(SPI_REG_DIAG_STATUS_2) & 0x10))
        {
            gStepper_Obj.StallState = SUCCESS;  // Tell user stall learning was a success
//            if(gStepper_Obj.REVERSE_ON_STALL) LearnStallTHOnce = true;
        }
        else if(gStepper_Obj.StallState < SUCCESS)
            gStepper_Obj.StallState = BEGIN;
    }
    if(gStepper_Obj.StallState == DISABLED && last_stall_state != DISABLED)
    {
        gStepper_Obj.REVERSE_ON_STALL = false;
        SPI_writeRegister(SPI_REG_CTRL5, DIS_STL_MASK);
    }
}

// Stepper State machine
// This is used to move the motor
void MotorStateMachine()
{
    // update the SteppingRateTMR if required
    // reduce the number of calculations
    if (UpdateSteppingRateTMR == true)
    {
        SteppingRateTMR = SMCLK_FREQ/SteppingRate;  // calculate new value
        UpdateSteppingRateTMR = false;              // clear flag
    }
    switch (gStepper_Obj.AccelerateState)
    {
    case (IDLE):
            if(gStepper_Obj.RECIPROCATION == false && gDRV8889_Obj.nSleep == 1 && gDRV8889_Obj.gDRVOFF == 0)
            {
                gStepper_Obj.MOTOR_CMD_ACTIVE1 = false;
                gStepper_Obj.MOTOR_CMD_ACTIVE2 = false;
                gStepper_Obj.MOTOR_CMD_ACTIVE3 = false;
                gStepper_Obj.MANUAL_ADVANCE = false;           // clear the flags once stops
                MoveSteps = false;                  // clear the MoveSteps flag here
                // reset the timers if the last command was run motor
                if (LastCommandWasRunSteps == true)
                {
                    TACTL &= ~MC_3;                     // Stop timerA
                    TBCTL &= ~MC_3;                     // Stop timerB
                }

                if(RecoverFromStall == true) // after stall fault, wait a little bit, change direction, them begin driving again
                {
                    __delay_cycles(10000000);
                    if(gStepper_Obj.gDIR == 1)                // change direction
                    {
                        gStepper_Obj.gDIR = 0;
                    }
                    else gStepper_Obj.gDIR = 1;

                    RecoverFromStall = false;
                    SPI_writeRegister(SPI_REG_CTRL4, CLR_FLT_MASK);     // Clear fault
                    SPI_writeRegister(SPI_REG_CTRL5, DIS_STL_MASK);     // disable stall detection during acceleration
                    gStepper_Obj.AccelerateState = ACCEL;
                }
                if((gStepper_Obj.RECIPROCATION == true && gStepper_Obj.MANUAL_ADVANCE == true))
                {
                    __delay_cycles(10000000);   // pause here if reciprocating
                }
            }
    break;
    case (ACCEL):
        {
        // do nothing in ACCEL mode, wait for external input to change state
        }
    break;
    case (DECEL):
        {
        // do nothing in DECEL mode, wait for external input to change state
        }
    break;
    case (RUN):
        {
            if(gStepper_Obj.StallState > DISABLED && gStepper_Obj.StallState < LEARNING)
            {
                last_stall_state = gStepper_Obj.StallState;
                gStepper_Obj.StallState = LEARNING;                                // set StallState to LEARNING
                stall_learn_number_of_steps = 0;                                   // set step counter to 0
                SPI_writeRegister(SPI_REG_CTRL5, 0x28);                            // set STL_LRN bit to begin learning stall, set STL_REP bit to report stall on nFAULT
            }
            if(gStepper_Obj.StallState == READY4STALL && read_nFAULT_pin == 0)
            {
                if((SPI_readRegister(SPI_REG_DIAG_STATUS_2) & 0x10))
                {
                    last_stall_state = gStepper_Obj.StallState;
                    gStepper_Obj.StallState = SUCCESS;  // Tell user stall learning was a success
                }
            }
        }
    break;
    case (DECEL2NEWTARGET):
        {
        // do nothing in DECEL mode, wait for external input to change state
        }
    }
}

void DetermineMicrostepping(char CTRL3Data)
{
    MICROSTEP_MODE_BITS = CTRL3Data & 0x0F;
    switch(CTRL3Data & 0x0F)
    {
        case (0x00):    // Full step (2-phase excitation) with 100% current
        case (0x01):    // Full step (2-phase excitation) with 71% current
            {
                stall_learn_microstep_divider = 2;
            }
        break;
        case (0x02):    // Non-circular 1/2 step
        case (0x03):    // 1/2 step
            {
                stall_learn_microstep_divider = 3;
            }
        break;
        default:    // 1/4 step to 1/32 step
            {
                stall_learn_microstep_divider = CTRL3Data & 0x0F;
            }
    }
}
